# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the faces service implementation for v1."""

from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest
import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.conversions import (
    to_area,
)
from ..base.faces import GRPCFacesService
from .conversions import (
    build_grpc_id,
    from_angle_to_grpc_quantity,
    from_grpc_curve_to_curve,
    from_grpc_direction_to_unit_vector,
    from_grpc_point_to_point3d,
    from_grpc_surface_to_surface,
    from_length_to_grpc_quantity,
    from_line_to_grpc_line,
    from_parameter_to_grpc_quantity,
    from_point3d_to_grpc_point,
    from_unit_vector_to_grpc_direction,
    serialize_tracked_command_response,
)


class GRPCFacesServiceV1(GRPCFacesService):
    """Faces service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    faces service. It is specifically designed for the v1 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2_grpc import FaceStub
        from ansys.api.discovery.v1.operations.edit_pb2_grpc import EditStub

        self.stub = FaceStub(channel)
        self.edit_stub = EditStub(channel)

    @protect_grpc
    def get_surface(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetSurface(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "surface": from_grpc_surface_to_surface(response.surface, kwargs["surface_type"]),
        }

    @protect_grpc
    def get_box_uv(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetBoxUV(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "uv_box": {
                "u": (
                    response.start_u.value_in_geometry_units,
                    response.end_u.value_in_geometry_units,
                ),
                "v": (
                    response.start_v.value_in_geometry_units,
                    response.end_v.value_in_geometry_units,
                ),
            }
        }

    @protect_grpc
    def get_area(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetArea(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {"area": to_area(response.area.value_in_geometry_units)}

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetEdges(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in response.edges
            ]
        }

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetVertices(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "vertices": [
                {
                    "id": vertex.id.id,
                    "position": from_grpc_point_to_point3d(vertex.position),
                }
                for vertex in response.vertices
            ]
        }

    @protect_grpc
    def get_loops(self, **kwargs) -> dict:  # noqa: D102
        from ..base.conversions import to_distance

        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetLoops(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "loops": [
                {
                    "type": loop.type,
                    "length": to_distance(loop.length.value_in_geometry_units).value,
                    "min_corner": from_grpc_point_to_point3d(loop.bounding_box.min),
                    "max_corner": from_grpc_point_to_point3d(loop.bounding_box.max),
                    "edges": [edge for edge in loop.edges],
                }
                for loop in response.loops
            ]
        }

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        response = self.stub.GetColor(request=request)

        # Return the response - formatted as a dictionary
        color = response.colors.get(kwargs["id"], "")
        return {"color": color}

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            GetBoundingBoxRequest,
            GetBoundingBoxRequestData,
        )

        # Create the request to the proper method depending on tight tolerenace
        if kwargs.get("tight"):
            request = GetBoundingBoxRequest(
                request_data=[
                    GetBoundingBoxRequestData(
                        id=build_grpc_id(kwargs["id"]),
                        tight_tolerance=kwargs.get("tight", False),
                    )
                ]
            )

            resp = self.stub.GetTightBoundingBox(request).response_data[0]
        else:
            request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])
            resp = self.stub.GetBoundingBox(request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "min_corner": from_grpc_point_to_point3d(resp.box.min),
            "max_corner": from_grpc_point_to_point3d(resp.box.max),
            "center": from_grpc_point_to_point3d(resp.box.center),
        }

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            SetColorRequest,
            SetColorRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetColorRequest(
            request_data=[
                SetColorRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    color=kwargs["color"],
                )
            ]
        )
        # Call the gRPC service
        response = self.stub.SetColor(request=request)

        # Return the response - formatted as a dictionary
        return {"success": len(response.successfully_set_ids) == 1}

    @protect_grpc
    def get_normal(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2 import (
            GetNormalRequest,
            GetNormalRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetNormalRequest(
            request_data=[
                GetNormalRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    u=kwargs["u"],
                    v=kwargs["v"],
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.GetNormal(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "normal": from_grpc_direction_to_unit_vector(response.direction),
        }

    @protect_grpc
    def evaluate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2 import (
            EvaluateRequest,
            EvaluateRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = EvaluateRequest(
            request_data=[
                EvaluateRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    u=kwargs["u"],
                    v=kwargs["v"],
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.Evaluate(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "point": from_grpc_point_to_point3d(response.point),
        }

    @protect_grpc
    def create_iso_parametric_curve(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2 import (
            CreateIsoParamCurvesRequest,
            CreateIsoParamCurvesRequestData,
        )

        from ansys.geometry.core.shapes.parameterization import Interval

        from ..base.conversions import to_distance

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateIsoParamCurvesRequest(
            request_data=[
                CreateIsoParamCurvesRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    u_dir_curve=kwargs["use_u_param"],
                    proportion=from_parameter_to_grpc_quantity(kwargs["parameter"]),
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.CreateIsoParamCurves(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "curves": [
                {
                    "geometry": from_grpc_curve_to_curve(curve.curve),
                    "start": from_grpc_point_to_point3d(curve.start),
                    "end": from_grpc_point_to_point3d(curve.end),
                    "interval": Interval(curve.interval_start, curve.interval_end),
                    "length": to_distance(curve.length.value_in_geometry_units).value,
                }
                for curve in response.curves
            ]
        }

    @protect_grpc
    def extrude_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ExtrudeFacesRequest,
            ExtrudeFacesRequestData,
        )

        # Assign direction
        direction = (
            None
            if kwargs["direction"] is None
            else from_unit_vector_to_grpc_direction(kwargs["direction"])
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtrudeFacesRequest(
            request_data=[
                ExtrudeFacesRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    distance=from_length_to_grpc_quantity(kwargs["distance"]),
                    direction=direction,
                    extrude_type=kwargs["extrude_type"].value,
                    pull_symmetric=kwargs["pull_symmetric"],
                    offset_mode=kwargs["offset_mode"].value,
                    copy=kwargs["copy"],
                    force_do_as_extrude=kwargs["force_do_as_extrude"],
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.ExtrudeFaces(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": tracked_response.get("success"),
            "created_bodies": [
                body.get("id") for body in tracked_response.get("created_bodies", [])
            ],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def extrude_faces_up_to(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ExtrudeFacesUpToRequest,
            ExtrudeFacesUpToRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtrudeFacesUpToRequest(
            request_data=[
                ExtrudeFacesUpToRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    up_to_selection_id=build_grpc_id(kwargs["up_to_selection_id"]),
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    extrude_type=kwargs["extrude_type"].value,
                    pull_symmetric=kwargs["pull_symmetric"],
                    offset_mode=kwargs["offset_mode"].value,
                    copy=kwargs["copy"],
                    force_do_as_extrude=kwargs["force_do_as_extrude"],
                    seed_point=from_point3d_to_grpc_point(kwargs["seed_point"]),
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.ExtrudeFacesUpTo(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": tracked_response.get("success"),
            "created_bodies": [
                body.get("id").id for body in tracked_response.get("created_bodies", [])
            ],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def offset_faces_set_radius(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            OffsetFacesSetRadiusRequest,
            OffsetFacesSetRadiusRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = OffsetFacesSetRadiusRequest(
            request_data=[
                OffsetFacesSetRadiusRequestData(
                    faces_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    radius=from_length_to_grpc_quantity(kwargs["radius"]),
                    offset_mode=kwargs["offset_mode"].value,
                    copy=kwargs["copy"],
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.OffsetFacesSetRadius(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def revolve_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            RevolveFacesRequest,
            RevolveFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesRequest(
            request_data=[
                RevolveFacesRequestData(
                    selection_id=[build_grpc_id(id) for id in kwargs["selection_ids"]],
                    axis=from_line_to_grpc_line(kwargs["axis"]),
                    angle=from_angle_to_grpc_quantity(kwargs["angle"]),
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.RevolveFaces(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": tracked_response.get("success"),
            "created_bodies": [
                body.get("id").id for body in tracked_response.get("created_bodies")
            ],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def revolve_faces_up_to(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            RevolveFacesUpToRequest,
            RevolveFacesUpToRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesUpToRequest(
            request_data=[
                RevolveFacesUpToRequestData(
                    selection_ids=[build_grpc_id(id) for id in kwargs["selection_ids"]],
                    up_to_selection_id=build_grpc_id(kwargs["up_to_selection_id"]),
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    axis=from_line_to_grpc_line(kwargs["axis"]),
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.RevolveFacesUpTo(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": tracked_response.get("success"),
            "created_bodies": [
                body.get("id").id for body in tracked_response.get("created_bodies")
            ],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def revolve_faces_by_helix(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            RevolveFacesByHelixRequest,
            RevolveFacesByHelixRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = RevolveFacesByHelixRequest(
            request_data=[
                RevolveFacesByHelixRequestData(
                    selection_ids=[build_grpc_id(id) for id in kwargs["selection_ids"]],
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    axis=from_line_to_grpc_line(kwargs["axis"]),
                    height=from_length_to_grpc_quantity(kwargs["height"]),
                    pitch=from_length_to_grpc_quantity(kwargs["pitch"]),
                    taper_angle=from_angle_to_grpc_quantity(kwargs["taper_angle"]),
                    right_handed=kwargs["right_handed"],
                    both_sides=kwargs["both_sides"],
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.RevolveFacesByHelix(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": tracked_response.get("success"),
            "created_bodies": [body.get("id") for body in tracked_response.get("created_bodies")],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def replace_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2 import ReplaceFaceRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ReplaceFaceRequest(
            target_selection_ids=[build_grpc_id(id) for id in kwargs["target_ids"]],
            replacement_selection_ids=[build_grpc_id(id) for id in kwargs["replacement_ids"]],
        )

        # Call the gRPC service
        response = self.stub.Replace(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def thicken_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ThickenFacesRequest,
            ThickenFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = ThickenFacesRequest(
            request_data=[
                ThickenFacesRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    value=from_length_to_grpc_quantity(kwargs["thickness"]),
                    extrude_type=kwargs["extrude_type"].value,
                    pull_symmetric=kwargs["pull_symmetric"],
                    select_direction=kwargs["select_direction"],
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.ThickenFaces(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def draft_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            DraftFacesRequest,
            DraftFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = DraftFacesRequest(
            request_data=[
                DraftFacesRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    reference_ids=[build_grpc_id(id) for id in kwargs["reference_face_ids"]],
                    draft_side=kwargs["draft_side"].value,
                    draft_angle=from_angle_to_grpc_quantity(kwargs["angle"]),
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC server
        response = self.edit_stub.DraftFaces(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the drafted faces
        return {
            "created_faces": [face.id for face in response.created_faces],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def get_round_info(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["face_id"])])

        # Call the gRPC service
        response = self.stub.GetRoundInfo(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "along_u": response.along_u,
            "radius": response.radius.value_in_geometry_units,
        }

    @protect_grpc
    def offset_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            OffsetFacesRequest,
            OffsetFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = OffsetFacesRequest(
            request_data=[
                OffsetFacesRequestData(
                    face_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    offset=from_length_to_grpc_quantity(kwargs["distance"]),
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    extrude_type=kwargs["extrude_type"].value,
                )
            ]
        )

        # Call the gRPC service and serialize the response
        response = self.edit_stub.OffsetFaces(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "results": [face.get("id").id for face in tracked_response.get("created_faces")],
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def setup_offset_relationship(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            FaceOffsetRequest,
            FaceOffsetRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FaceOffsetRequest(
            request_data=[
                FaceOffsetRequestData(
                    id1=build_grpc_id(kwargs["face1_id"]),
                    id2=build_grpc_id(kwargs["face2_id"]),
                    set_baselines=kwargs["set_baselines"],
                    process_adjacent_faces=kwargs["process_adjacent_faces"],
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.FaceOffset(request=request)
        tracked_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_response": tracked_response,
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the unsupported service implementation for v1."""

from ansys.api.discovery.v1.commands.unsupported_pb2 import SetExportIdData, SetExportIdRequest
import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.unsupported import GRPCUnsupportedService
from .conversions import build_grpc_id


class GRPCUnsupportedServiceV1(GRPCUnsupportedService):
    """Unsupported service for gRPC communication with the Geometry server.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.commands.unsupported_pb2_grpc import UnsupportedStub

        self.stub = UnsupportedStub(channel)

    @protect_grpc
    def get_import_id_map(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commands.unsupported_pb2 import GetImportIdMapRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetImportIdMapRequest(type=kwargs["id_type"].value)

        # Call the gRPC service
        response = self.stub.GetImportIdMap(request)

        # Return the response - formatted as a dictionary
        return {
            "id_map": {k: v.id for k, v in response.id_map.items()},  # dict[str, str]
            "id_type": kwargs["id_type"],
        }

    @protect_grpc
    def set_export_ids(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = SetExportIdRequest(
            export_data=[
                SetExportIdData(
                    moniker_id=build_grpc_id(data.moniker),
                    id=data.value,
                    type=data.id_type.value,
                )
                for data in kwargs["export_data"]
            ]
        )

        # Call the gRPC service
        _ = self.stub.SetExportId(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_single_export_id(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = SetExportIdRequest(
            export_data=[
                SetExportIdData(
                    moniker_id=build_grpc_id(kwargs["export_data"].moniker),
                    id=kwargs["export_data"].value,
                    type=kwargs["export_data"].id_type.value,
                )
            ]
        )

        # Call the gRPC service
        _ = self.stub.SetExportId(request)

        # Return the response - formatted as a dictionary
        return {}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Unsupported functions for the PyAnsys Geometry library."""

from dataclasses import dataclass
from enum import Enum, unique
from typing import TYPE_CHECKING

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.misc.auxiliary import get_all_bodies_from_design
from ansys.geometry.core.misc.checks import (
    min_backend_version,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face
    from ansys.geometry.core.modeler import Modeler


@unique
class PersistentIdType(Enum):
    """Type of persistent id."""

    PNAME = 1
    PRIME_ID = 700


@dataclass
class ExportIdData:
    """Data for exporting persistent ids."""

    moniker: str
    id_type: PersistentIdType
    value: str


class UnsupportedCommands:
    """Provides unsupported commands for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        gRPC client to use for the geometry commands.
    modeler : Modeler
        Modeler instance to use for the geometry commands.
    _internal_use : bool, optional
        Internal flag to prevent direct instantiation by users.
        This parameter is for internal use only.

    Raises
    ------
    GeometryRuntimeError
        If the class is instantiated directly by users instead of through the modeler.

    Notes
    -----
    This class should not be instantiated directly. Use
    ``modeler.unsupported`` instead.

    """

    def __init__(self, grpc_client: GrpcClient, modeler: "Modeler", _internal_use: bool = False):
        """Initialize an instance of the ``UnsupportedCommands`` class."""
        if not _internal_use:
            raise GeometryRuntimeError(
                "UnsupportedCommands should not be instantiated directly. "
                "Use 'modeler.unsupported' to access unsupported commands."
            )
        self._grpc_client = grpc_client
        self.__id_map = {}
        self.__modeler = modeler
        self.__current_design = modeler.get_active_design()

    @min_backend_version(25, 2, 0)
    def __fill_imported_id_map(self, id_type: PersistentIdType) -> None:
        """Populate the persistent id map for caching.

        This cache should be cleared on design change.

        Parameters
        ----------
        id_type : PersistentIdType
            Type of id.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        result = self._grpc_client.services.unsupported.get_import_id_map(id_type=id_type)
        self.__id_map[id_type] = result.get("id_map", {})

    def __is_occurrence(self, master: str, occ: str) -> bool:
        """Determine if the master is the master of the occurrence.

        Parameters
        ----------
        master : str
            Master moniker.
        occ : str
            Occurrence moniker.

        Returns
        -------
        bool
            ``True`` if the master is the master of the occurrence.

        """
        master_id = occ.split("/")[-1]
        return master == master_id

    def __get_moniker_from_import_id(self, id_type: PersistentIdType, import_id: str) -> str | None:
        """Look up the moniker from the id map.

        Parameters
        ----------
        id_type : PersistentIdType
            Type of id.
        import_id : str
            Persistent id.

        Returns
        -------
        str | None
            Moniker associated with the id or None

        Notes
        -----
        This checks if the design has changed and clears the cache if it has.
        """
        if (
            self.__current_design != self.__modeler.get_active_design()
            or id_type not in self.__id_map
        ):
            self.__fill_imported_id_map(id_type)
        moniker = self.__id_map[id_type].get(import_id, None)
        return moniker

    @min_backend_version(25, 2, 0)
    def set_export_id(self, moniker: str, id_type: PersistentIdType, value: str) -> None:
        """Set the persistent id for the moniker.

        Parameters
        ----------
        moniker : str
            Moniker to set the id for.
        id_type : PersistentIdType
            Type of id.
        value : str
            Id to set.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        self._grpc_client.services.unsupported.set_single_export_id(
            export_data=ExportIdData(moniker=moniker, id_type=id_type, value=value),
        )
        self.__id_map = {}

    @min_backend_version(26, 1, 0)
    def set_multiple_export_ids(
        self,
        export_data: list[ExportIdData],
    ) -> None:
        """Set multiple persistent ids for the monikers.

        Parameters
        ----------
        export_data : list[ExportIdData]
            List of export data containing monikers, id types, and values.

        Warnings
        --------
        This method is only available starting on Ansys release 26R1.
        """
        # Call the gRPC service
        _ = self._grpc_client.services.unsupported.set_export_ids(export_data=export_data)
        self.__id_map = {}

    def get_body_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Body"]:
        """Get all body occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
            Persistent id
        id_type : PersistentIdType
            Type of id

        Returns
        -------
        list[Body]
            List of body occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            body
            for body in get_all_bodies_from_design(design)
            if self.__is_occurrence(moniker, body.id)
        ]

    def get_face_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Face"]:
        """Get all face occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
            Persistent id.
        id_type : PersistentIdType
            Type of id.

        Returns
        -------
        list[Face]
            List of face occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            face
            for body in get_all_bodies_from_design(design)
            for face in body.faces
            if self.__is_occurrence(moniker, face.id)
        ]

    def get_edge_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Edge"]:
        """Get all edge occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
           Persistent id.
        id_type : PersistentIdType
           Type of id.

        Returns
        -------
        list[Edge]
           List of edge occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            edge
            for body in get_all_bodies_from_design(design)
            for edge in body.edges
            if self.__is_occurrence(moniker, edge.id)
        ]

    @min_backend_version(27, 1, 0)
    def start_tracking(self) -> None:
        """Start a tracking segment for the current design.

        Warnings
        --------
        This method is only available starting on Ansys release 27R1.
        """
        design = self.__modeler.get_active_design()

        # Disable automatic tracking and start tracking for the document
        self._grpc_client.services.admin.set_automatic_tracking_state(
            enabled=False, design_id=design.id
        )
        self._grpc_client.services.admin.get_tracker(design_id=design.id)

    @min_backend_version(27, 1, 0)
    def stop_tracking(self) -> dict:
        """Stop the current tracking segment for the current design and return the tracked changes.

        Returns
        -------
        dict
            Dictionary containing the tracked changes.

        Warnings
        --------
        This method is only available starting on Ansys release 27R1.
        """
        design = self.__modeler.get_active_design()
        changes = self._grpc_client.services.admin.get_tracker_changes(design_id=design.id)

        # Re-enable automatic tracking
        self._grpc_client.services.admin.set_automatic_tracking_state(
            enabled=True, design_id=design.id
        )

        return changes



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for connecting to a local Geometry Service Docker container."""

from enum import Enum
from functools import wraps
import os
from pathlib import Path
from typing import Optional

from beartype import beartype as check_input_types

try:
    from docker.client import DockerClient
    from docker.errors import APIError, ContainerError, ImageNotFound
    from docker.models.containers import Container

    _HAS_DOCKER = True
except ModuleNotFoundError:  # pragma: no cover
    _HAS_DOCKER = False

from ansys.tools.common.cyberchannel import verify_transport_mode

import ansys.geometry.core.connection.defaults as pygeom_defaults
from ansys.geometry.core.logger import LOG


def _docker_python_available(func):
    """Check whether Docker is installed as a Python package.

    This function works as a decorator.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not _HAS_DOCKER:  # pragma: no cover
            raise ModuleNotFoundError(
                "The package 'docker' is required to use the LocalDockerInstance class."
            )
        else:
            return func(*args, **kwargs)

    return wrapper


class GeometryContainers(Enum):
    """Provides an enum holding the available Geometry services."""

    CORE_WINDOWS_LATEST = 0, "windows", "core-windows-latest"
    CORE_LINUX_LATEST = 1, "linux", "core-linux-latest"
    CORE_WINDOWS_LATEST_UNSTABLE = 2, "windows", "core-windows-latest-unstable"
    CORE_LINUX_LATEST_UNSTABLE = 3, "linux", "core-linux-latest-unstable"
    # TO BE REMOVED at some point... - START
    WINDOWS_LATEST = 4, "windows", "windows-latest"
    WINDOWS_LATEST_UNSTABLE = 5, "windows", "windows-latest-unstable"
    # TO BE REMOVED at some point... - END
    WINDOWS_24_1 = 6, "windows", "windows-24.1"
    WINDOWS_24_2 = 7, "windows", "windows-24.2"
    WINDOWS_25_1 = 8, "windows", "windows-25.1"
    WINDOWS_25_2 = 9, "windows", "windows-25.2"
    CORE_WINDOWS_25_2 = 10, "windows", "core-windows-25.2"
    CORE_LINUX_25_2 = 11, "linux", "core-linux-25.2"
    CORE_WINDOWS_26_1 = 10, "windows", "core-windows-26.1"
    CORE_LINUX_26_1 = 11, "linux", "core-linux-26.1"
    CORE_WINDOWS_27_1 = 12, "windows", "core-windows-27.1"
    CORE_LINUX_27_1 = 13, "linux", "core-linux-27.1"


class LocalDockerInstance:
    """Instantiates a Geometry service as a local Docker container.

    By default, if a container with the Geometry service already exists at the given port,
    PyAnsys Geometry connects to it. Otherwise, PyAnsys Geometry tries to launch its own service.

    Parameters
    ----------
    port : int, optional
        Localhost port to deploy the Geometry service on or the
        the ``Modeler`` interface to connect to (if it is already deployed). By default,
        the value is the one for the ``DEFAULT_PORT`` connection parameter.
    connect_to_existing_service : bool, default: True
        Whether the ``Modeler`` interface should connect to a Geometry
        service already deployed at the specified port.
    restart_if_existing_service : bool, default: False
        Whether the Geometry service (which is already running)
        should be restarted when attempting connection.
    name : str or None, default: None
        Name of the Docker container to deploy. The default is ``None``,
        in which case Docker assigns it a random name.
    image : GeometryContainers | str | None, default: None
        The Geometry service Docker image to deploy. This can be either:

        * A ``GeometryContainers`` enum value for predefined images
        * A string containing a custom Docker image name (e.g., myregistry.com/my-geometry:tag)

        The default is ``None``, in which case the ``LocalDockerInstance`` class identifies
        the OS of your Docker engine and deploys the latest version of the Geometry service
        for that OS.
    transport_mode : str | None
        Transport mode selected, by default `None` and thus it will be selected
        for you based on the connection criteria. Options are: "insecure", "mtls"
    certs_dir : Path | str | None
        Directory to use for TLS certificates.
        By default `None` and thus search for the "ANSYS_GRPC_CERTIFICATES" environment variable.
        If not found, it will use the "certs" folder assuming it is in the current working
        directory.
    """

    __DOCKER_CLIENT__: "DockerClient" = None
    """Docker client class variable.

    Notes
    -----
    The default is ``None``, in which case lazy initialization is used.
    ``__DOCKER_CLIENT__`` is a class variable, meaning that it is
    the same variable for all instances of this class.
    """

    @staticmethod
    @_docker_python_available
    def docker_client() -> "DockerClient":
        """Get the initialized ``__DOCKER_CLIENT__`` object.

        Returns
        -------
        ~docker.client.DockerClient
            Initialized Docker client.

        Notes
        -----
        The ``LocalDockerInstance`` class performs a lazy initialization of the
        ``__DOCKER_CLIENT__`` class variable.
        """
        if not LocalDockerInstance.__DOCKER_CLIENT__:
            LocalDockerInstance.__DOCKER_CLIENT__ = DockerClient.from_env()

        return LocalDockerInstance.__DOCKER_CLIENT__

    @staticmethod
    @_docker_python_available
    def is_docker_installed() -> bool:
        """Check a local installation of Docker engine is available and running.

        Returns
        -------
        bool
            ``True`` if Docker engine is available and running, ``False`` otherwise.
        """
        # When trying to ping, this throws an error if the
        # server is not available
        try:
            return LocalDockerInstance.docker_client().ping()
        except APIError:  # pragma: no cover
            return False

    @check_input_types
    @_docker_python_available
    def __init__(
        self,
        port: int = pygeom_defaults.DEFAULT_PORT,
        connect_to_existing_service: bool = True,
        restart_if_existing_service: bool = False,
        name: str | None = None,
        image: GeometryContainers | str | None = None,
        transport_mode: str | None = None,
        certs_dir: Path | str | None = None,
    ) -> None:
        """``LocalDockerInstance`` constructor."""
        # Initialize instance variables
        self._container: Container = None
        self._existed_previously: bool = False
        self._transport_mode: str | None = transport_mode

        # Check the port availability
        port_available, cont = self._check_port_availability(port)

        # If a service was already deployed... check if it is a Geometry service
        # in case you want to connect to it
        if cont and connect_to_existing_service and self._is_cont_geom_service(cont):
            # Now, check if a restart of the service is required
            if restart_if_existing_service:
                LOG.warning(f"Restarting service already running at port {port}...")
                cont.restart()

            # Finally, store the container
            self._container = cont
            self._existed_previously = True
            # If no transport mode was provided, try to extract it from the existing
            # container command line arguments
            if transport_mode is None:
                cmd_args = cont.attrs["Config"].get("Cmd", [])
                for arg in cmd_args:
                    if "--transport-mode=" in arg:
                        self._transport_mode = arg.split("=")[1]
                        break
                # If still None, default to "mtls"
                if self._transport_mode is None:
                    self._transport_mode = "mtls"
            return

        # At this stage, confirm that you have to deploy our own Geometry service.
        #
        # First, check if the port is available... otherwise raise error
        if port_available:
            self._deploy_container(
                port=port,
                name=name,
                image=image,
                transport_mode=transport_mode,
                certs_dir=certs_dir,
            )
        else:
            raise RuntimeError(f"Geometry service cannot be deployed on port {port}")

    def _check_port_availability(self, port: int) -> tuple[bool, Optional["Container"]]:
        """Check whether the requested port is available for deployment.

        Returns
        -------
        tuple[bool, Container]
            Tuple with a Boolean indicating whether the port is available.
            If the port is not available, the container deployed at that port is also returned.
        """
        # First, check if there is a container already running at that port
        for cont in self.docker_client().containers.list():
            for _, ports_shared in cont.attrs["NetworkSettings"]["Ports"].items():
                # If no ports are shared, continue looping
                if not ports_shared:
                    continue

                # Check shared ports otherwise
                for port_shared in ports_shared:
                    if int(port_shared["HostPort"]) == port:
                        LOG.warning(f"Service is already running at port {port}...")
                        return (False, cont)

        # If you reached here, just return default values
        return (True, None)

    def _is_cont_geom_service(self, cont: "Container") -> bool:
        """Check a provided ``Container`` object is a Geometry service container.

        Parameters
        ----------
        cont : Container
            Container deployed to check.

        Returns
        -------
        bool
            Returns ``True`` if the container provided is a Geometry service.
            Otherwise, ``False``.
        """
        # If one of the tags matches a Geometry service tag --> Return True
        for tag in cont.image.tags:
            for geom_services in GeometryContainers:
                if (
                    tag
                    == f"{pygeom_defaults.GEOMETRY_SERVICE_DOCKER_IMAGE}:{geom_services.value[2]}"
                ):
                    return True

        # If you have reached this point, the image is not a Geometry service
        return False  # pragma: no cover

    def _deploy_container(
        self,
        port: int,
        name: str | None,
        image: GeometryContainers | str | None,
        transport_mode: str | None,
        certs_dir: Path | str | None,
    ) -> None:
        """Handle the deployment of a Geometry service.

        Parameters
        ----------
        port : int
            Port under which the service should be deployed.
        name : str or None, optional
            Name given to the deployed container. If ``None``, Docker will provide
            an arbitrary name.
        image : GeometryContainers | str | None
            Geometry service Docker container image to be used. This can be either:

            * A ``GeometryContainers`` enum value for predefined images
            * A string containing a custom Docker image name
            * ``None`` - the latest container version matching the Docker engine OS is used
        transport_mode : str | None
            Transport mode selected, by default `None` and thus it will be selected
            for you based on the connection criteria. Options are: "insecure", "mtls"
        certs_dir : Path | str | None
            Directory to use for TLS certificates.
            By default `None` and thus search for the "ANSYS_GRPC_CERTIFICATES" environment
            variable. If not found, it will use the "certs" folder assuming it is in the
            current working directory.

        Raises
        ------
        RuntimeError
            In case the Geometry service cannot be launched.
        """
        # First get the Docker Engine OS
        docker_os = self.docker_client().info()["OSType"]

        # Determine the final image to use
        final_image_name = None

        if isinstance(image, str):
            # Custom image name provided as string - use it directly
            final_image_name = image
            LOG.info(f"Using custom Docker image: {final_image_name}")
        elif image is None:
            # If no image is provided, default to whatever Docker engine OS
            # your system is running on... Latest images are used because they are the
            # ones defined first in the GeometryContainers enum class.
            for geom_service in GeometryContainers:
                if geom_service.value[1] == docker_os:
                    image = geom_service
                    break

            # If image is still None, this means it cannot be deployed on your OS
            if image is None:  # pragma: no cover
                raise RuntimeError(f"Geometry service cannot be launched on {docker_os}")

            final_image_name = f"{pygeom_defaults.GEOMETRY_SERVICE_DOCKER_IMAGE}:{image.value[2]}"
        else:
            # GeometryContainers enum provided
            # Check if the requested image is compatible with the existing OS
            if image.value[1] != docker_os:  # pragma: no cover
                raise RuntimeError(f"Geometry service cannot be launched on {docker_os}")

            final_image_name = f"{pygeom_defaults.GEOMETRY_SERVICE_DOCKER_IMAGE}:{image.value[2]}"

        # At this point, you are can deploy the Geometry service.
        #
        # Check if the license server env variable is available
        license_server = os.getenv("ANSRV_GEO_LICENSE_SERVER", None)
        if not license_server:  # pragma: no cover
            raise RuntimeError(
                "No license server provided... Store its value under the following env variable: ANSRV_GEO_LICENSE_SERVER."  # noqa: E501
            )

        # Verify the transport mode
        if transport_mode:
            verify_transport_mode(transport_mode, "remote")
        else:
            # Default to "mtls" if possible
            transport_mode = "mtls"
            LOG.info(f"No transport mode provided. Defaulting to '{transport_mode}'")

        # Create the shared volume if mtls is selected
        volumes = None
        if transport_mode == "mtls":
            # Share the certificates directory if needed
            if certs_dir is None:
                certs_dir_env = os.getenv("ANSYS_GRPC_CERTIFICATES", None)
                if certs_dir_env is not None:
                    certs_dir = certs_dir_env
                else:
                    certs_dir = Path.cwd() / "certs"

            if not Path(certs_dir).is_dir():  # pragma: no cover
                raise RuntimeError(
                    "Transport mode 'mtls' was selected, but the expected"
                    f" certificates directory does not exist: {certs_dir}"
                )

            # Determine the container-side path based on the image OS
            # For custom images, we need to infer from docker engine OS
            container_os = image.value[1] if isinstance(image, GeometryContainers) else docker_os
            container_cert_path = "/certs" if container_os == "linux" else "C:/certs"

            volumes = {
                str(Path(certs_dir).resolve()): {
                    "bind": container_cert_path,
                    "mode": "ro",
                }
            }

        # Try to deploy it
        try:
            # Determine the container-side cert path for environment variables
            container_os = image.value[1] if isinstance(image, GeometryContainers) else docker_os
            container_cert_path = "/certs" if container_os == "linux" else "C:/certs"

            container: Container = self.docker_client().containers.run(
                image=final_image_name,
                detach=True,
                auto_remove=True,
                network_mode="bridge" if docker_os == "linux" else "nat",
                name=name,
                ports={"50051/tcp": port},
                volumes=volumes,
                environment={
                    "LICENSE_SERVER": license_server,
                    "LOG_LEVEL": os.getenv("ANSRV_GEO_LOG_LEVEL", 2),
                    "ENABLE_TRACE": os.getenv("ANSRV_GEO_ENABLE_TRACE", 0),
                    "USE_DEBUG_MODE": os.getenv("ANSRV_GEO_USE_DEBUG_MODE", 0),
                    "SERVER_ENDPOINT": "50051@0.0.0.0",
                    "ANSYS_GRPC_CERTIFICATES": container_cert_path,
                },
                command=f"--transport-mode={transport_mode}",
            )
        except ImageNotFound:  # pragma: no cover
            raise RuntimeError(
                f"Geometry service Docker image '{final_image_name}' not found. Download it first to your machine."  # noqa: E501
            )
        except (ContainerError, APIError) as err:
            raise RuntimeError(
                f"Geometry service Docker image error when initialized. Error: {err.explanation}"
            )
        except Exception as ex:  # pragma: no cover
            raise RuntimeError(
                f"Geometry service Docker image could not be launched. Error: {str(ex)}"
            )

        # If the deployment went fine, this means that you have deployed the service.
        self._container = container
        self._existed_previously = False
        self._transport_mode = transport_mode

    @property
    def container(self) -> "Container":
        """Docker container object that hosts the deployed Geometry service."""
        return self._container

    @property
    def existed_previously(self) -> bool:
        """Flag indicating whether the container previously existed.

        Returns ``False`` if the Geometry service was effectively
        deployed by this class or ``True`` if it already existed.
        """
        return self._existed_previously

    @property
    def transport_mode(self) -> str:
        """Transport mode used by the Docker container.

        Returns
        -------
        str
            Transport mode used by the Docker container. If no transport mode
            was specified during initialization, it defaults to "mtls".
        """
        return self._transport_mode if self._transport_mode else "mtls"


def get_geometry_container_type(instance: LocalDockerInstance) -> GeometryContainers | None:
    """Provide back the ``GeometryContainers`` value.

    Parameters
    ----------
    instance : LocalDockerInstance
        The LocalDockerInstance object.

    Returns
    -------
    GeometryContainers or None
        The GeometryContainer value corresponding to the previous image or None
        if not match.

    Notes
    -----
    This method returns the first hit on the available tags.
    """
    for tag in instance.container.image.tags:
        for geom_services in GeometryContainers:
            if tag == f"{pygeom_defaults.GEOMETRY_SERVICE_DOCKER_IMAGE}:{geom_services.value[2]}":
                return geom_services

    return None



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the Commands Script service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.commands_script import GRPCCommandsScriptService


class GRPCCommandsScriptServiceV1(GRPCCommandsScriptService):
    """Commands Script service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    Commands Script service. It is specifically designed for the v1 version
    of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.commands.script_pb2_grpc import ScriptStub

        self.stub = ScriptStub(channel)

    @protect_grpc
    def run_script_file(self, **kwargs) -> dict:  # noqa: D102
        from pathlib import Path
        from typing import Generator

        from ansys.api.discovery.v1.commands.script_pb2 import RunScriptFileRequest

        import ansys.geometry.core.connection.defaults as pygeom_defaults

        def request_generator(
            script_path: Path,
            file_path: str,
            script_args: dict,
            api_version: int | None,
        ) -> Generator[RunScriptFileRequest, None, None]:
            """Generate requests for streaming file upload."""
            msg_buffer = 5 * 1024  # 5KB - for additional message data
            if pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer < 0:  # pragma: no cover
                raise ValueError("MAX_MESSAGE_LENGTH is too small for file upload.")

            chunk_size = pygeom_defaults.MAX_MESSAGE_LENGTH - msg_buffer
            with Path.open(script_path, "rb") as file:
                while chunk := file.read(chunk_size):
                    test_req = RunScriptFileRequest(
                        script_path=str(file_path),
                        script_args=script_args,
                        api_version=api_version,
                        data=chunk,
                    )
                    yield test_req

        # Call the gRPC service
        response = self.stub.RunScriptFile(
            request_generator(
                script_path=kwargs["script_path"],
                file_path=kwargs["original_path"],
                script_args=kwargs["script_args"],
                api_version=kwargs["api_version"],
            )
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.command_response.success,
            "message": response.command_response.message,
            "values": None if not response.values else dict(response.values),
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a body."""

from abc import ABC, abstractmethod
from collections.abc import Iterable
from enum import Enum, unique
from functools import wraps
from typing import TYPE_CHECKING, Union

from beartype import beartype as check_input_types
import matplotlib.colors as mcolors
from pint import Quantity

import ansys.geometry.core as pyansys_geom
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.designer.edge import CurveType, Edge
from ansys.geometry.core.designer.face import Face, SurfaceType
from ansys.geometry.core.designer.vertex import Vertex
from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.math.bbox import BoundingBox
from ansys.geometry.core.math.constants import IDENTITY_MATRIX44
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.auxiliary import (
    DEFAULT_COLOR,
    convert_color_to_hex,
    convert_opacity_to_hex,
    get_bodies_from_ids,
    get_design_from_body,
)
from ansys.geometry.core.misc.checks import (
    check_nurbs_compatibility,
    check_type,
    check_type_all_elements_in_iterable,
    ensure_design_is_active,
    graphics_required,
    min_backend_version,
)
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.options import TessellationOptions
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.sketch.sketch import Sketch
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from pyvista import MultiBlock, PolyData
    from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet, vtkPolyData

    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.selection import NamedSelection

# TODO: Temporary fix for boolean operations
# This is a temporary fix for the boolean operations issue. The issue is that the
# boolean operations are not working as expected with command-based operations. The
# goal is to fix this issue in the future.
# https://github.com/ansys/pyansys-geometry/issues/1733
__TEMPORARY_BOOL_OPS_FIX__ = (99, 0, 0)


@unique
class MidSurfaceOffsetType(Enum):
    """Provides values for mid-surface offsets supported."""

    MIDDLE = 0
    TOP = 1
    BOTTOM = 2
    VARIABLE = 3
    CUSTOM = 4


@unique
class CollisionType(Enum):
    """Provides values for collision types between bodies."""

    NONE = 0
    TOUCH = 1
    INTERSECT = 2
    CONTAINED = 3
    CONTAINEDTOUCH = 4


@unique
class FillStyle(Enum):
    """Provides values for fill styles supported."""

    DEFAULT = 0
    OPAQUE = 1
    TRANSPARENT = 2


class IBody(ABC):
    """Defines the common methods for a body, providing the abstract body interface.

    Both the ``MasterBody`` class and ``Body`` class both inherit from the ``IBody``
    class. All child classes must implement all abstract methods.
    """

    @abstractmethod
    def id(self) -> str:
        """Get the ID of the body as a string."""
        return

    @abstractmethod
    def name(self) -> str:
        """Get the name of the body."""
        return

    @abstractmethod
    def set_name(self, str) -> None:
        """Set the name of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        return

    @abstractmethod
    def fill_style(self) -> FillStyle:
        """Get the fill style of the body."""
        return

    @abstractmethod
    def set_fill_style(self, fill_style: FillStyle) -> None:
        """Set the fill style of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        return

    @abstractmethod
    def is_suppressed(self) -> bool:
        """Get the body suppression state.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        return

    @abstractmethod
    def set_suppressed(self, suppressed: bool) -> None:
        """Set the body suppression state.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        return

    @abstractmethod
    def color(self) -> str:
        """Get the color of the body."""
        return

    @abstractmethod
    def opacity(self) -> float:
        """Get the float value of the opacity for the body."""
        return

    @abstractmethod
    def set_color(
        self, color: str | tuple[float, float, float] | tuple[float, float, float, float]
    ) -> None:
        """Set the color of the body.

        Parameters
        ----------
        color : str | tuple[float, float, float] | tuple[float, float, float, float]
            Color to set the body to. This can be a string representing a color name
            or a tuple of RGB values in the range [0, 1] (RGBA) or [0, 255] (pure RGB).

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        return

    @abstractmethod
    def faces(self) -> list[Face]:
        """Get a list of all faces within the body.

        Returns
        -------
        list[Face]
        """
        return

    @abstractmethod
    def edges(self) -> list[Edge]:
        """Get a list of all edges within the body.

        Returns
        -------
        list[Edge]
        """
        return

    @abstractmethod
    def vertices(self) -> list[Vertex]:
        """Get a list of all vertices within the body.

        Returns
        -------
        list[Vertex]
        """
        return

    @abstractmethod
    def is_alive(self) -> bool:
        """Check if the body is still alive and has not been deleted."""
        return

    @abstractmethod
    def is_surface(self) -> bool:
        """Check if the body is a planar body."""
        return

    @abstractmethod
    def surface_thickness(self) -> Quantity | None:
        """Get the surface thickness of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface thickness.
        """
        return

    @abstractmethod
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:
        """Get the surface offset type of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface offset.
        """
        return

    @abstractmethod
    def volume(self) -> Quantity:
        """Calculate the volume of the body.

        Notes
        -----
        When dealing with a planar surface, a value of ``0`` is returned as a volume.
        """
        return

    @abstractmethod
    def material(self) -> Material:
        """Get the assigned material of the body.

        Returns
        -------
        Material
            Material assigned to the body.
        """
        return

    @abstractmethod
    def bounding_box(self) -> BoundingBox:
        """Get the bounding box of the body.

        Returns
        -------
        BoundingBox
            Bounding box of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        return

    @abstractmethod
    def get_bounding_box(self, tight: bool = False) -> BoundingBox:
        """Get the bounding box of the body.

        Parameters
        ----------
        tight : bool, default: False
            Whether to use a tight tolerance when calculating the bounding box.

        Returns
        -------
        BoundingBox
            Bounding box of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 27R1.
        """
        return

    @abstractmethod
    def assign_material(self, material: Material) -> None:
        """Assign a material against the active design.

        Parameters
        ----------
        material : Material
            Source material data.
        """
        return

    @abstractmethod
    def get_assigned_material(self) -> Material:
        """Get the assigned material of the body.

        Returns
        -------
        Material
            Material assigned to the body.
        """
        return

    @abstractmethod
    def remove_assigned_material(self) -> None:
        """Remove the material assigned to the body."""
        return

    @abstractmethod
    def add_midsurface_thickness(self, thickness: Distance | Quantity | Real) -> None:
        """Add a mid-surface thickness to a surface body.

        Parameters
        ----------
        thickness : Distance | Quantity | Real
            Thickness to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface thickness assignment.
        """
        return

    @abstractmethod
    def add_midsurface_offset(self, offset: "MidSurfaceOffsetType") -> None:
        """Add a mid-surface offset to a surface body.

        Parameters
        ----------
        offset_type : MidSurfaceOffsetType
            Surface offset to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface offset assignment.
        """
        return

    @abstractmethod
    def imprint_curves(self, faces: list[Face], sketch: Sketch) -> tuple[list[Edge], list[Face]]:
        """Imprint all specified geometries onto specified faces of the body.

        Parameters
        ----------
        faces: list[Face]
            list of faces to imprint the curves of the sketch onto.
        sketch: Sketch
            All curves to imprint on the faces.

        Returns
        -------
        tuple[list[Edge], list[Face]]
            All impacted edges and faces from the imprint operation.
        """
        return

    @abstractmethod
    def project_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        """Project all specified geometries onto the body.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Returns
        -------
        list[Face]
            All faces from the project curves operation.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.
        """
        return

    @abstractmethod
    def imprint_projected_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:  # noqa: D102
        """Project and imprint specified geometries onto the body.

        This method combines the ``project_curves()`` and ``imprint_curves()`` method into
        one method. It has higher performance than calling them back-to-back when dealing
        with many curves. Because it is a specialized function, this method only returns
        the faces (and not the edges) from the imprint operation.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Returns
        -------
        list[Face]
            All imprinted faces from the operation.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.
        """
        return

    @abstractmethod
    def translate(self, direction: UnitVector3D, distance: Quantity | Distance | Real) -> None:
        """Translate the body in a specified direction and distance.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the translation.
        distance: ~pint.Quantity | Distance | Real
            Distance (magnitude) of the translation.
        """
        return

    @abstractmethod
    def rotate(
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        """Rotate the geometry body around the specified axis by a given angle.

        Parameters
        ----------
        axis_origin: Point3D
            Origin of the rotational axis.
        axis_direction: UnitVector3D
            The axis of rotation.
        angle: ~pint.Quantity | Angle | Real
            Angle (magnitude) of the rotation.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        return

    @abstractmethod
    def scale(self, value: Real) -> None:
        """Scale the geometry body by the given value.

        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        value: Real
            Value to scale the body by.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        return

    @abstractmethod
    def map(self, frame: Frame) -> None:
        """Map the geometry body to the new specified frame.

        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        frame: Frame
            Structure defining the orientation of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        return

    @abstractmethod
    def mirror(self, plane: Plane) -> None:
        """Mirror the geometry body across the specified plane.

        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        plane: Plane
            Represents the mirror.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        return

    @abstractmethod
    def get_collision(self, body: "Body") -> CollisionType:
        """Get the collision state between bodies.

        Parameters
        ----------
        body: Body
            Object that the collision state is checked with.

        Returns
        -------
        CollisionType
            Enum that defines the collision state between bodies.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        return

    @abstractmethod
    def copy(self, parent: "Component", name: str = None) -> "Body":
        """Create a copy of the body under the specified parent.

        Parameters
        ----------
        parent: Component
            Parent component to place the new body under within the design assembly.
        name: str
            Name to give the new body.

        Returns
        -------
        Body
            Copy of the body.
        """
        return

    @abstractmethod
    def get_named_selections(self) -> list["NamedSelection"]:
        """Get the named selections associated with the body.

        Returns
        -------
        list[NamedSelection]
            List of named selections associated with the body.
        """
        return

    @abstractmethod
    def get_raw_tessellation(
        self,
        transform: Matrix44 = IDENTITY_MATRIX44,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> dict:
        """Tessellate the body and return the raw tessellation data.

        Parameters
        ----------
        transform : Matrix44, default: IDENTITY_MATRIX44
            A transformation matrix to apply to the tessellation.
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.
        reset_cache : bool, default: False
            Whether to reset the tessellation cache and re-request the tessellation
            from the server.
        include_faces : bool, default: True
            Whether to include face tessellation data in the output.
        include_edges : bool, default: False
            Whether to include edge tessellation data in the output.

        Returns
        -------
        dict
            Dictionary with face and edge IDs as keys and face and edge tessellation data as values.
        """

    @abstractmethod
    def tessellate(
        self,
        merge: bool = False,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> Union["PolyData", "MultiBlock"]:
        """Tessellate the body and return the geometry as triangles.

        Parameters
        ----------
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``False`` (default), the
            number of triangles are preserved and only the topology is merged.
            When ``True``, the individual faces of the tessellation are merged.
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.
        reset_cache : bool, default: False
            Whether to reset the tessellation cache and re-request the tessellation
            from the server.
        include_faces : bool, default: True
            Whether to include face tessellation data in the output.
        include_edges : bool, default: False
            Whether to include edge tessellation data in the output.

        Returns
        -------
        ~pyvista.PolyData, ~pyvista.MultiBlock
            Merged :class:`pyvista.PolyData` if ``merge=True`` or a composite dataset.

        Examples
        --------
        Extrude a box centered at the origin to create a rectangular body and
        tessellate it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> my_comp = design.add_component("my-comp")
        >>> body = my_comp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> blocks = body.tessellate()
        >>> blocks
        >>> MultiBlock(0x7F94EC757460)
             N Blocks:	6
             X Bounds:	0.000, 4.000
             Y Bounds:	-1.000, 0.000
             Z Bounds:	-0.500, 4.500

        Merge the body:

        >>> mesh = body.tessellate(merge=True)
        >>> mesh
        PolyData (0x7f94ec75f3a0)
          N Cells:	12
          N Points:	24
          X Bounds:	0.000e+00, 4.000e+00
          Y Bounds:	-1.000e+00, 0.000e+00
          Z Bounds:	-5.000e-01, 4.500e+00
          N Arrays:	0
        """
        return

    @abstractmethod
    def get_vtk_tessellation(
        self,
        merge: bool = False,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
        _raw_tessellation: dict | None = None,
    ) -> Union["vtkPolyData", "vtkMultiBlockDataSet"]:
        """Build VTK objects from raw tessellation data.

        Parameters
        ----------
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``False`` (default), the
            number of triangles are preserved and only the topology is merged.
            When ``True``, the individual faces of the tessellation are merged.
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.
        reset_cache : bool, default: False
            Whether to reset the tessellation cache and re-request the tessellation
            from the server.
        include_faces : bool, default: True
            Whether to include face tessellation data in the output.
        include_edges : bool, default: False
            Whether to include edge tessellation data in the output.
        _raw_tessellation : dict | None, default: None
            Raw tessellation data with face and edge IDs as keys and tessellation data as
            values.. If ``None``, the method requests the raw tessellation data from the server.
            This parameter is intended for internal use only.

        Returns
        -------
        vtkPolyData, vtkMultiBlockDataSet
            Merged `vtkPolyData` if ``merge=True`` or a `vtkMultiBlockDataSet`.
        """
        return

    @abstractmethod
    def shell_body(self, offset: Distance | Quantity | Real) -> bool:
        """Shell the body to the thickness specified.

        Parameters
        ----------
        offset : Distance | Quantity | Real
            Shell thickness.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        return

    @abstractmethod
    def remove_faces(
        self, selection: Face | Iterable[Face], offset: Distance | Quantity | Real
    ) -> bool:
        """Shell by removing a given set of faces.

        Parameters
        ----------
        selection : Face | Iterable[Face]
            Face or faces to be removed.
        offset : Distance | Quantity | Real
            Shell thickness.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        return

    @abstractmethod
    def plot(
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        """Plot the body.

        Parameters
        ----------
        merge : bool, default: True
            Whether to merge the body into a single mesh. Performance improved when ``True``.
            When ``True`` (default), the individual faces of the tessellation are merged.
            When ``False``, the number of triangles are preserved and only the topology
            is merged.
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the
            ``ansys.tools.visualization_interface.USE_TRAME`` global setting is used.
        use_service_colors : bool, default: None
            Whether to use the colors assigned to the body in the service. The default
            is ``None``, in which case the ``ansys.geometry.core.USE_SERVICE_COLORS``
            global setting is used.
        **plotting_options : dict, default: None
            Keyword arguments for plotting. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Examples
        --------
        Extrude a box centered at the origin to create rectangular body and
        plot it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> mycomp = design.add_component("my-comp")
        >>> body = mycomp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> body.plot()

        Plot the body and color each face individually:

        >>> body.plot(multi_colors=True)
        """
        return

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Intersect two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to intersect with.
        keep_other : bool, default: False
            Whether to retain the intersected body or not.

        Raises
        ------
        ValueError
            If the bodies do not intersect.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the intersected body is retained.
        """
        return

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Subtract two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to subtract from the ``self`` parameter.
        keep_other : bool, default: False
            Whether to retain the subtracted body or not.

        Raises
        ------
        ValueError
            If the subtraction results in an empty (complete) subtraction.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the subtracted body is retained.
        """
        return

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Unite two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to unite with the ``self`` parameter.
        keep_other : bool, default: False
            Whether to retain the united body or not.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the united body is retained.
        """
        return

    def combine_merge(self, other: Union["Body", list["Body"]]) -> None:
        """Combine this body with another body or bodies, merging them into a single body.

        Parameters
        ----------
        other : Union[Body, list[Body]]
            The body or list of bodies to combine with this body.

        Notes
        -----
        The ``self`` parameter is directly modified, and the ``other`` bodies are consumed.
        """
        return

    def _combine_subtract(
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool = False,
        transfer_named_selections=True,
    ) -> None:
        """Subtract bodies from this body.

        Parameters
        ----------
        other : Union[Body, list[Body]]
            The body or list of bodies to combine with this body.
        keep_other : bool, default: False
            Whether to retain the other bodies or not.

        Warnings
        --------
        This is a specialized boolean operation that has the ability to transfer named
        selections. It may behave differently than the encouraged ``subtract()``.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the united body is retained.
        """
        return


class MasterBody(IBody):
    """Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    is_surface : bool, default: False
        Whether the master body is a surface or an 3D object (with volume). The default
        is ``False``, in which case the master body is a surface. When ``True``, the
        master body is a 3D object (with volume).
    """

    def __init__(
        self,
        id: str,
        name: str,
        grpc_client: GrpcClient,
        is_surface: bool = False,
    ):
        """Initialize the ``MasterBody`` class."""
        check_type(id, str)
        check_type(name, str)
        check_type(grpc_client, GrpcClient)
        check_type(is_surface, bool)

        self._id = id
        self._name = name
        self._grpc_client = grpc_client
        self._is_surface = is_surface
        self._surface_thickness = None
        self._surface_offset = None
        self._is_alive = True
        self._tessellation = None
        self._raw_tessellation = None
        self._fill_style = FillStyle.DEFAULT
        self._color = None

    def reset_tessellation_cache(func):  # noqa: N805
        """Decorate ``MasterBody`` methods that need tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "MasterBody", *args, **kwargs):
            self._tessellation = None
            self._raw_tessellation = None
            return func(self, *args, **kwargs)

        return wrapper

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @property
    def name(self) -> str:  # noqa: D102
        return self._name

    @name.setter
    def name(self, value: str):  # noqa: D102
        self.set_name(value)

    @property
    def fill_style(self) -> str:  # noqa: D102
        return self._fill_style

    @fill_style.setter
    def fill_style(self, value: FillStyle):  # noqa: D102
        self.set_fill_style(value)

    @property
    @min_backend_version(25, 2, 0)
    def is_suppressed(self) -> bool:  # noqa: D102
        response = self._grpc_client.services.bodies.is_suppressed(id=self.id)
        return response.get("result")

    @is_suppressed.setter
    def is_suppressed(self, value: bool):  # noqa: D102
        self.set_suppressed(value)

    @property
    def color(self) -> str:  # noqa: D102
        """Get the current color of the body."""
        if self._color is None and self.is_alive:
            # Assigning default value first
            self._color = DEFAULT_COLOR

            if self._grpc_client.backend_version < (25, 1, 0):  # pragma: no cover
                # Server does not support color retrieval before version 25.1.0
                self._grpc_client.log.warning(
                    "Server does not support color retrieval. Default value assigned..."
                )
            else:
                # Fetch color from the server if it's not cached
                color = self._grpc_client.services.bodies.get_color(id=self.id).get("color")
                if color:
                    self._color = mcolors.to_hex(color, keep_alpha=True)

        return self._color

    @property
    def opacity(self) -> float:  # noqa: D102
        opacity_hex = self.color[7:]
        return int(opacity_hex, 16) / 255 if opacity_hex else 1

    @color.setter
    def color(self, value: str | tuple[float, float, float]) -> None:  # noqa: D102
        self.set_color(value)

    @opacity.setter
    def opacity(self, value: float) -> None:  # noqa: D102
        self.set_opacity(value)

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._is_surface

    @property
    def surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._surface_thickness if self.is_surface else None

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset if self.is_surface else None

    @property
    def faces(self) -> list[Face]:  # noqa: D102
        return self._get_faces_from_id(self)

    def _get_faces_from_id(self, body: Union["Body", "MasterBody"]) -> list[Face]:
        """Retrieve faces from a body ID."""
        self._grpc_client.log.debug(f"Retrieving faces for body {body.id} from server.")
        response = self._grpc_client.services.bodies.get_faces(id=body.id)
        return [
            Face(
                face_resp.get("id"),
                SurfaceType(face_resp.get("surface_type")),
                body,
                self._grpc_client,
                face_resp.get("is_reversed"),
            )
            for face_resp in response.get("faces")
        ]

    @property
    def edges(self) -> list[Edge]:  # noqa: D102
        return self._get_edges_from_id(self)

    def _get_edges_from_id(self, body: Union["Body", "MasterBody"]) -> list[Edge]:
        """Retrieve edges from a body ID."""
        self._grpc_client.log.debug(f"Retrieving edges for body {body.id} from server.")
        response = self._grpc_client.services.bodies.get_edges(id=body.id)
        return [
            Edge(
                edge_resp.get("id"),
                CurveType(edge_resp.get("curve_type")),
                body,
                self._grpc_client,
                edge_resp.get("is_reversed"),
            )
            for edge_resp in response.get("edges")
        ]

    @property
    @min_backend_version(26, 1, 0)
    def vertices(self) -> list[Vertex]:  # noqa: D102
        return self._get_vertices_from_id(self)

    def _get_vertices_from_id(self, body: Union["Body", "MasterBody"]) -> list[Vertex]:
        """Retrieve vertices from a body ID."""
        self._grpc_client.log.debug(f"Retrieving vertices for body {body.id} from server.")
        response = self._grpc_client.services.bodies.get_vertices(id=body.id)

        return [
            Vertex(
                vertex_resp.get("id"),
                vertex_resp.get("position"),
                body,
            )
            for vertex_resp in response.get("vertices")
        ]

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    def volume(self) -> Quantity:  # noqa: D102
        if self.is_surface:
            self._grpc_client.log.debug("Dealing with planar surface. Returning 0 as the volume.")
            return Quantity(0, DEFAULT_UNITS.SERVER_VOLUME)
        else:
            self._grpc_client.log.debug(f"Retrieving volume for body {self.id} from server.")
            response = self._grpc_client.services.bodies.get_volume(id=self.id)
            return response.get("volume")

    @property
    def material(self) -> Material:  # noqa: D102
        return self.get_assigned_material()

    @material.setter
    def material(self, value: Material):  # noqa: D102
        self.assign_material(value)

    @property
    @min_backend_version(25, 2, 0)
    def bounding_box(self) -> BoundingBox:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving bounding box for body {self.id} from server.")
        response = self._grpc_client.services.bodies.get_bounding_box(id=self.id)
        return BoundingBox(
            min_corner=response.get("min"),
            max_corner=response.get("max"),
            center=response.get("center"),
        )

    @min_backend_version(27, 1, 0)
    def get_bounding_box(self, tight: bool = False) -> BoundingBox:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving bounding box for body {self.id} from server.")
        response = self._grpc_client.services.bodies.get_bounding_box(id=self.id, tight=tight)

        return BoundingBox(
            min_corner=response.get("min"),
            max_corner=response.get("max"),
            center=response.get("center"),
        )

    @check_input_types
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Assigning body {self.id} material {material.name}.")
        self._grpc_client.services.bodies.set_assigned_material(id=self.id, material=material)

    def get_assigned_material(self) -> Material:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving assigned material for body {self.id}.")
        response = self._grpc_client.services.bodies.get_assigned_material(id=self.id)
        return response.get("material")

    @min_backend_version(26, 1, 0)
    def remove_assigned_material(self) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Removing assigned material for body {self.id}.")
        self._grpc_client.services.bodies.remove_assigned_material(ids=[self.id])

    @check_input_types
    def add_midsurface_thickness(self, thickness: Distance | Quantity | Real) -> None:  # noqa: D102
        thickness = thickness if isinstance(thickness, Distance) else Distance(thickness)

        if self.is_surface:
            self._grpc_client.services.bodies.assign_midsurface_thickness(
                ids=[self.id], thickness=thickness
            )
            self._surface_thickness = thickness.value
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface thickness because it is not a surface. Ignoring request."  # noqa : E501
            )

    @check_input_types
    def add_midsurface_offset(self, offset: MidSurfaceOffsetType) -> None:  # noqa: D102
        if self.is_surface:
            self._grpc_client.services.bodies.assign_midsurface_offset(
                ids=[self.id], offset_type=offset
            )
            self._surface_offset = offset
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface offset because it is not a surface. Ignoring request."  # noqa : E501
            )

    @check_input_types
    def imprint_curves(  # noqa: D102
        self, faces: list[Face], sketch: Sketch
    ) -> tuple[list[Edge], list[Face]]:
        raise NotImplementedError(
            """
            imprint_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    def project_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        raise NotImplementedError(
            """
            project_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    def imprint_projected_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        raise NotImplementedError(
            """
            imprint_projected_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    @reset_tessellation_cache
    def translate(  # noqa: D102
        self, direction: UnitVector3D, distance: Quantity | Distance | Real
    ) -> None:
        distance = distance if isinstance(distance, Distance) else Distance(distance)
        self._grpc_client.log.debug(f"Translating body {self.id}.")
        self._grpc_client.services.bodies.translate(
            ids=[self.id], direction=direction, distance=distance
        )

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_name(  # noqa: D102
        self, name: str
    ) -> None:
        self._grpc_client.log.debug(f"Renaming body {self.id} from '{self.name}' to '{name}'.")
        self._grpc_client.services.bodies.set_name(id=self.id, name=name)
        self._name = name

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_fill_style(  # noqa: D102
        self, fill_style: FillStyle
    ) -> None:
        self._grpc_client.log.debug(f"Setting body fill style {self.id}.")
        self._grpc_client.services.bodies.set_fill_style(id=self.id, fill_style=fill_style)
        self._fill_style = fill_style

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_suppressed(  # noqa: D102
        self, suppressed: bool
    ) -> None:
        self._grpc_client.log.debug(f"Setting body {self.id}, as suppressed: {suppressed}.")
        self._grpc_client.services.bodies.set_suppressed(bodies=[self.id], is_suppressed=suppressed)

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_color(  # noqa: D102
        self, color: str | tuple[float, float, float] | tuple[float, float, float, float]
    ) -> None:
        self._grpc_client.log.debug(f"Setting body color of {self.id} to {color}.")
        color = convert_color_to_hex(color)
        self._grpc_client.services.bodies.set_color(id=self.id, color=color)
        self._color = color

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_opacity(self, opacity: float) -> None:
        """Set the opacity of the body.

        Warnings
        --------
        This method is only available starting on Ansys release 25R2.
        """
        self._grpc_client.log.debug(f"Setting body opacity of {self.id} to {opacity}.")
        opacity = convert_opacity_to_hex(opacity)
        new_color = self.color[0:7] + opacity
        self.set_color(new_color)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def rotate(  # noqa: D102
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        angle = angle if isinstance(angle, Angle) else Angle(angle)
        self._grpc_client.log.debug(f"Rotating body {self.id}.")
        self._grpc_client.services.bodies.rotate(
            id=self.id, axis_origin=axis_origin, axis_direction=axis_direction, angle=angle
        )

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def scale(self, value: Real) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Scaling body {self.id}.")
        self._grpc_client.services.bodies.scale(id=self.id, value=value)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def map(self, frame: Frame) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mapping body {self.id}.")
        self._grpc_client.services.bodies.map(id=self.id, frame=frame)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mirroring body {self.id}.")
        self._grpc_client.services.bodies.mirror(id=self.id, plane=plane)

    @min_backend_version(24, 2, 0)
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        self._grpc_client.log.debug(f"Get collision between body {self.id} and body {body.id}.")
        response = self._grpc_client.services.bodies.get_collision(id=self.id, other_id=body.id)
        return CollisionType(response.get("collision_type"))

    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        raise NotImplementedError(
            "Copy method is not implemented on the MasterBody. Call this method on a body instead."
        )

    def get_named_selections(self) -> list["NamedSelection"]:  # noqa: D102
        raise NotImplementedError(
            """
            get_named_selections is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @min_backend_version(26, 1, 0)
    def get_raw_tessellation(  # noqa: D102
        self,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> dict:
        if not self.is_alive:
            return {}

        self._grpc_client.log.debug(f"Requesting tessellation for body {self.id}.")

        # cache tessellation
        if not self._raw_tessellation or reset_cache:
            response = self._grpc_client.services.bodies.get_full_tessellation(
                id=self.id,
                options=tess_options,
                raw_data=True,
                include_faces=include_faces,
                include_edges=include_edges,
            )

            self._raw_tessellation = response.get("tessellation")

        return self._raw_tessellation

    @graphics_required
    def tessellate(  # noqa: D102
        self,
        merge: bool = False,
        transform: Matrix44 = IDENTITY_MATRIX44,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> Union["PolyData", "MultiBlock"]:
        # lazy import here to improve initial module load time
        import pyvista as pv

        if not self.is_alive:
            return pv.PolyData() if merge else pv.MultiBlock()

        # If the server does not support tessellation options, ignore them
        if tess_options is not None and self._grpc_client.backend_version < (25, 2, 0):
            self._grpc_client.log.warning(
                "Tessellation options are not supported by server"
                f" version {self._grpc_client.backend_version}. Ignoring options."
            )
            tess_options = None
        if include_edges and self._grpc_client.backend_version < (26, 1, 0):
            self._grpc_client.log.warning(
                "Edge tessellation is not supported by server"
                f" version {self._grpc_client.backend_version}. Ignoring request."
            )
            include_edges = False

        self._grpc_client.log.debug(f"Requesting tessellation for body {self.id}.")

        # cache tessellation
        if not self._tessellation or reset_cache:
            if self._grpc_client.backend_version > (25, 2, 0):
                response = self._grpc_client.services.bodies.get_full_tessellation(
                    id=self.id,
                    options=tess_options,
                    raw_data=False,
                    include_faces=include_faces,
                    include_edges=include_edges,
                )
            elif tess_options is not None:
                response = self._grpc_client.services.bodies.get_tesellation_with_options(
                    id=self.id, options=tess_options, raw_data=False
                )
            else:
                response = self._grpc_client.services.bodies.get_tesellation(
                    id=self.id, backend_version=self._grpc_client.backend_version, raw_data=False
                )

            self._tessellation = response.get("tessellation")

        if transform == IDENTITY_MATRIX44:
            pdata = list(self._tessellation.values())
        else:
            pdata = [
                tess.transform(transform, inplace=False) for tess in self._tessellation.values()
            ]
        comp = pv.MultiBlock(pdata)

        if merge:
            ugrid = comp.combine()
            return pv.PolyData(var_inp=ugrid.points, faces=ugrid.cells)
        else:
            return comp

    @graphics_required
    def get_vtk_tessellation(  # noqa: D102
        self,
        merge: bool = False,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
        _raw_tessellation: dict | None = None,
    ) -> Union["vtkPolyData", "vtkMultiBlockDataSet"]:
        # lazy import here to improve initial module load time
        import numpy as np
        import vtk
        from vtkmodules.vtkCommonCore import vtkPoints
        from vtkmodules.vtkCommonDataModel import vtkCellArray, vtkMultiBlockDataSet, vtkPolyData

        def _process_vertices(vertex_data: list[float]):
            """Process vertex data and create VTK points."""
            points = vtkPoints()
            # Each vertex has 3 coordinates (x, y, z) and vertex_data
            # is a flat list, so we reshape it accordingly
            point_data = np.array(vertex_data).reshape(-1, 3)
            for point in point_data:
                points.InsertNextPoint(point[0], point[1], point[2])
            return points

        def _process_faces(face_data: list[int]):
            """Process face data and create VTK cells."""
            cells = vtkCellArray()
            cell_array = np.array(face_data)

            # Algorithm to read the cell array:
            # The first integer indicates the number of points that define the face.
            # The next 'count' integers are the point indices.
            # This pattern repeats for all faces.
            if len(cell_array) > 0:
                i = 0
                while i < len(cell_array):
                    if i < len(cell_array):
                        count = int(cell_array[i])
                        if i + count < len(cell_array):
                            cells.InsertNextCell(count)
                            for j in range(count):
                                cells.InsertCellPoint(int(cell_array[i + 1 + j]))
                            i += count + 1
                        else:
                            break
                    else:
                        break
            return cells

        def _process_edges(vertex_data: list[float]):
            """Process edge data and create VTK line cells."""
            line_cells = vtkCellArray()

            # Each line is defined by two consecutive vertices, and
            # each vertex has 3 coordinates (x, y, z)... so the number of lines
            # is total number of vertices divided by 3, minus 1
            n_lines = len(vertex_data) // 3 - 1
            if n_lines > 0:
                for i in range(n_lines):
                    line_cells.InsertNextCell(2)
                    line_cells.InsertCellPoint(i)
                    line_cells.InsertCellPoint(i + 1)
            return line_cells

        def _create_polydata_from_tess_data(tess_data: dict):
            """Create a VTK PolyData object from tessellation data."""
            if not tess_data or len(tess_data.get("vertices", [])) == 0:
                return None

            polydata = vtkPolyData()
            vertex_data = tess_data["vertices"]

            # Set points
            points = _process_vertices(vertex_data)
            polydata.SetPoints(points)

            # Set face cells
            if "faces" in tess_data:
                cells = _process_faces(tess_data["faces"])
                polydata.SetPolys(cells)

            # Set edge lines
            if include_edges and tess_data.get("is_edge", False):
                line_cells = _process_edges(vertex_data)
                polydata.SetLines(line_cells)

            return polydata if polydata.GetNumberOfPoints() > 0 else None

        def _merge_vtk_objects(
            vtk_objects: list[vtkPolyData], merge: bool
        ) -> Union[vtkPolyData, vtkMultiBlockDataSet]:
            """Merge VTK objects into a single PolyData or MultiBlock."""
            if not vtk_objects:  # pragma: no cover
                self._grpc_client.log.warning("No valid tessellation data to convert.")
                return vtkPolyData() if merge else vtkMultiBlockDataSet()

            if merge:
                append_filter = vtk.vtkAppendPolyData()
                for vtk_obj in vtk_objects:
                    append_filter.AddInputData(vtk_obj)
                append_filter.Update()
                return append_filter.GetOutput()
            else:
                multiblock = vtkMultiBlockDataSet()
                multiblock.SetNumberOfBlocks(len(vtk_objects))
                for i, vtk_obj in enumerate(vtk_objects):
                    multiblock.SetBlock(i, vtk_obj)
                return multiblock

        ###### Main method logic #######
        # Use provided raw tessellation data or fetch it
        _raw_tessellation = _raw_tessellation or self.get_raw_tessellation(
            tess_options=tess_options,
            reset_cache=reset_cache,
            include_faces=include_faces,
            include_edges=include_edges,
        )

        # Check if tessellation data is available
        if not _raw_tessellation:
            self._grpc_client.log.warning("No tessellation data available.")
            return vtkPolyData() if merge else vtkMultiBlockDataSet()

        # Convert raw tessellation data to VTK PolyData objects
        vtk_objects = []
        for _, tess_data in _raw_tessellation.items():
            polydata = _create_polydata_from_tess_data(tess_data)
            if polydata:
                vtk_objects.append(polydata)

        # Merge VTK objects if required
        return _merge_vtk_objects(vtk_objects, merge)

    @reset_tessellation_cache
    @check_input_types
    @min_backend_version(25, 2, 0)
    def shell_body(self, offset: Distance | Quantity | Real) -> bool:  # noqa: D102
        self._grpc_client.log.debug(f"Shelling body {self.id} to offset {offset}.")

        offset = offset if isinstance(offset, Distance) else Distance(offset)
        result = self._grpc_client.services.bodies.shell(
            id=self.id,
            offset=offset,
        )

        if result.get("success") is False:
            self._grpc_client.log.warning(f"Failed to shell body {self.id}.")

        return result.get("success")

    @reset_tessellation_cache
    @check_input_types
    @min_backend_version(25, 2, 0)
    def remove_faces(  # noqa: D102
        self, selection: Face | Iterable[Face], offset: Distance | Quantity | Real
    ) -> bool:
        selection: list[Face] = selection if isinstance(selection, Iterable) else [selection]
        check_type_all_elements_in_iterable(selection, Face)

        # check if faces belong to this body
        for face in selection:
            if face.body.id != self.id:
                raise ValueError(f"Face {face.id} does not belong to body {self.id}.")

        self._grpc_client.log.debug(f"Removing faces to shell body {self.id}.")

        offset = offset if isinstance(offset, Distance) else Distance(offset)
        result = self._grpc_client.services.bodies.remove_faces(
            face_ids=[face.id for face in selection],
            offset=offset,
        )

        if result.get("success") is False:
            self._grpc_client.log.warning(f"Failed to remove faces from body {self.id}.")

        return result.get("success")

    @min_backend_version(25, 2, 0)
    @check_input_types
    def combine_merge(self, other: Union["Body", list["Body"]]) -> None:  # noqa: D102
        other = other if isinstance(other, list) else [other]
        check_type_all_elements_in_iterable(other, Body)

        self._grpc_client.log.debug(f"Combining and merging to body {self.id}.")
        self._grpc_client.services.bodies.combine_merge(
            body_ids=[self.id] + [body.id for body in other]
        )

    def _combine_subtract(  # noqa: D102
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool = False,
        transfer_named_selections=True,
    ) -> None:
        raise NotImplementedError(
            "MasterBody does not implement combine_subtract. Call this method on a body instead."
        )

    def plot(  # noqa: D102
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        raise NotImplementedError(
            "MasterBody does not implement plot methods. Call this method on a body instead."
        )

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def __repr__(self) -> str:
        """Represent the master body as a string."""
        lines = [f"ansys.geometry.core.designer.MasterBody {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")

        return "\n".join(lines)


class Body(IBody):
    """Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    template : MasterBody
        Master body that this body is an occurrence of.
    """

    def __init__(self, id, name, parent_component: "Component", template: MasterBody) -> None:
        """Initialize the ``Body`` class."""
        self._id = id
        self._name = name
        self._parent_component = parent_component
        self._template = template
        self._grpc_client = template._grpc_client

    def reset_tessellation_cache(func):  # noqa: N805
        """Decorate ``Body`` methods that require a tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "Body", *args, **kwargs):
            self._reset_tessellation_cache()
            return func(self, *args, **kwargs)

        return wrapper

    def _reset_tessellation_cache(self):  # noqa: N805
        """Reset the cached tessellation for a body."""
        self._template._tessellation = None
        self._template._raw_tessellation = None
        # if this reference is stale, reset the real cache in the part
        # this gets the matching id master body in the part
        master_in_part = next(
            (
                b
                for b in self.parent_component._master_component.part.bodies
                if b.id == self._template.id
            ),
            None,
        )
        if master_in_part is not None:
            master_in_part._tessellation = None
            master_in_part._raw_tessellation = None

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @property
    def name(self) -> str:  # noqa: D102
        return self._template.name

    @name.setter
    def name(self, value: str):  # noqa: D102
        self._template.name = value

    @property
    def fill_style(self) -> str:  # noqa: D102
        return self._template.fill_style

    @fill_style.setter
    def fill_style(self, fill_style: FillStyle) -> str:  # noqa: D102
        self._template.fill_style = fill_style

    @property
    def is_suppressed(self) -> bool:  # noqa: D102
        return self._template.is_suppressed

    @is_suppressed.setter
    def is_suppressed(self, suppressed: bool):  # noqa: D102
        self._template.is_suppressed = suppressed

    @property
    def color(self) -> str:  # noqa: D102
        return self._template.color

    @property
    def opacity(self) -> float:  # noqa: D102
        return self._template.opacity

    @color.setter
    def color(self, color: str | tuple[float, float, float]) -> None:  # noqa: D102
        return self._template.set_color(color)

    @opacity.setter
    def opacity(self, opacity: float) -> None:  # noqa: D102
        return self._template.set_opacity(opacity)

    @property
    def parent_component(self) -> "Component":  # noqa: D102
        return self._parent_component

    @property
    @ensure_design_is_active
    def faces(self) -> list[Face]:  # noqa: D102
        return self._template._get_faces_from_id(self)

    @property
    @ensure_design_is_active
    def edges(self) -> list[Edge]:  # noqa: D102
        return self._template._get_edges_from_id(self)

    @property
    @ensure_design_is_active
    def vertices(self) -> list[Vertex]:  # noqa: D102
        return self._template._get_vertices_from_id(self)

    @property
    def _is_alive(self) -> bool:  # noqa: D102
        return self._template.is_alive

    @_is_alive.setter
    def _is_alive(self, value: bool):  # noqa: D102
        self._template._is_alive = value

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._template.is_surface

    @property
    def _surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._template.surface_thickness

    @_surface_thickness.setter
    def _surface_thickness(self, value):  # noqa: D102
        self._template._surface_thickness = value

    @property
    def surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._surface_thickness

    @property
    def _surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._template._surface_offset

    @_surface_offset.setter
    def _surface_offset(self, value: "MidSurfaceOffsetType"):  # noqa: D102
        self._template._surface_offset = value

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset

    @property
    @ensure_design_is_active
    def volume(self) -> Quantity:  # noqa: D102
        return self._template.volume

    @property
    @ensure_design_is_active
    def material(self) -> Material:  # noqa: D102
        return self._template.material

    @material.setter
    def material(self, value: Material):  # noqa: D102
        self._template.material = value

    @property
    def bounding_box(self) -> BoundingBox:  # noqa: D102
        self._template._grpc_client.log.debug(
            f"Retrieving bounding box for body {self.id} from server."
        )
        response = self._template._grpc_client.services.bodies.get_bounding_box(id=self.id)
        return BoundingBox(
            min_corner=response.get("min"),
            max_corner=response.get("max"),
            center=response.get("center"),
        )

    @min_backend_version(27, 1, 0)
    def get_bounding_box(self, tight: bool = False) -> BoundingBox:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving bounding box for body {self.id} from server.")
        response = self._grpc_client.services.bodies.get_bounding_box(id=self.id, tight=tight)

        return BoundingBox(
            min_corner=response.get("min"),
            max_corner=response.get("max"),
            center=response.get("center"),
        )

    @ensure_design_is_active
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._template.assign_material(material)

    @ensure_design_is_active
    def get_assigned_material(self) -> Material:  # noqa: D102
        return self._template.get_assigned_material()

    @ensure_design_is_active
    def remove_assigned_material(self):  # noqa: D102
        self._template.remove_assigned_material()

    @ensure_design_is_active
    def add_midsurface_thickness(self, thickness: Quantity) -> None:  # noqa: D102
        self._template.add_midsurface_thickness(thickness)

    @ensure_design_is_active
    def add_midsurface_offset(  # noqa: D102
        self, offset: "MidSurfaceOffsetType"
    ) -> None:
        self._template.add_midsurface_offset(offset)

    @ensure_design_is_active
    def imprint_curves(  # noqa: D102
        self, faces: list[Face], sketch: Sketch = None, trimmed_curves: list[TrimmedCurve] = None
    ) -> tuple[list[Edge], list[Face]]:
        if sketch is None and self._template._grpc_client.backend_version < (25, 2, 0):
            raise ValueError(
                "A sketch must be provided for imprinting when using API versions below 25.2.0."
            )

        if sketch is None and trimmed_curves is None:
            raise ValueError("Either a sketch or edges must be provided for imprinting.")

        check_nurbs_compatibility(
            self._grpc_client.backend_version, sketch=sketch, curves=trimmed_curves
        )

        # Verify that each of the faces provided are part of this body
        body_faces = self.faces
        for provided_face in faces:
            if not any(provided_face.id == body_face.id for body_face in body_faces):
                raise ValueError(f"Face with ID {provided_face.id} is not part of this body.")

        self._template._grpc_client.log.debug(
            f"Imprinting curves on {self.id} for faces {[face.id for face in faces]}."
        )

        imprint_response = self._template._grpc_client.services.bodies.imprint_curves(
            id=self.id,
            sketch=sketch,
            face_ids=[face.id for face in faces],
            tc=trimmed_curves,
        )

        new_edges = [
            Edge(
                edge.get("id"),
                CurveType(edge.get("curve_type")),
                self,
                self._template._grpc_client,
            )
            for edge in imprint_response.get("edges")
        ]

        new_faces = [
            Face(
                face.get("id"),
                SurfaceType(face.get("surface_type")),
                self,
                self._template._grpc_client,
            )
            for face in imprint_response.get("faces")
        ]

        return (new_edges, new_faces)

    @ensure_design_is_active
    def project_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        check_nurbs_compatibility(self._grpc_client.backend_version, sketch=sketch)

        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        project_response = self._template._grpc_client.services.bodies.project_curves(
            id=self.id,
            sketch=sketch,
            direction=direction,
            closest_face=closest_face,
            only_one_curve=only_one_curve,
        )

        projected_faces = [
            Face(
                face.get("id"),
                SurfaceType(face.get("surface_type")),
                self,
                self._template._grpc_client,
            )
            for face in project_response.get("faces")
        ]

        return projected_faces

    @check_input_types
    @ensure_design_is_active
    def imprint_projected_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        check_nurbs_compatibility(self._grpc_client.backend_version, sketch=sketch)

        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        response = self._template._grpc_client.services.bodies.imprint_projected_curves(
            id=self.id,
            direction=direction,
            sketch=sketch,
            closest_face=closest_face,
            only_one_curve=only_one_curve,
        )

        imprinted_faces = [
            Face(
                face.get("id"),
                SurfaceType(face.get("surface_type")),
                self,
                self._template._grpc_client,
            )
            for face in response.get("faces")
        ]

        return imprinted_faces

    @ensure_design_is_active
    def set_name(self, name: str) -> None:  # noqa: D102
        return self._template.set_name(name)

    @ensure_design_is_active
    def set_fill_style(self, fill_style: FillStyle) -> None:  # noqa: D102
        return self._template.set_fill_style(fill_style)

    @ensure_design_is_active
    def set_suppressed(self, suppressed: bool) -> None:  # noqa: D102
        return self._template.set_suppressed(suppressed)

    @ensure_design_is_active
    def set_color(  # noqa: D102
        self, color: str | tuple[float, float, float] | tuple[float, float, float, float]
    ) -> None:
        return self._template.set_color(color)

    @ensure_design_is_active
    def translate(  # noqa: D102
        self, direction: UnitVector3D, distance: Quantity | Distance | Real
    ) -> None:
        return self._template.translate(direction, distance)

    @ensure_design_is_active
    def rotate(  # noqa: D102
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        return self._template.rotate(axis_origin, axis_direction, angle)

    @ensure_design_is_active
    def scale(self, value: Real) -> None:  # noqa: D102
        return self._template.scale(value)

    @ensure_design_is_active
    def map(self, frame: Frame) -> None:  # noqa: D102
        return self._template.map(frame)

    @ensure_design_is_active
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        return self._template.mirror(plane)

    @ensure_design_is_active
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        return self._template.get_collision(body)

    @ensure_design_is_active
    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        from ansys.geometry.core.designer.component import Component

        # Check input types
        check_type(parent, Component)
        copy_name = self.name if name is None else name
        check_type(copy_name, str)

        self._template._grpc_client.log.debug(f"Copying body {self.id}.")
        response = self._template._grpc_client.services.bodies.copy(
            id=self.id, parent_id=parent.id, name=copy_name
        )

        # Assign the new body to its specified parent (and return the new body)
        tb = MasterBody(
            response.get("master_id"),
            copy_name,
            self._template._grpc_client,
            is_surface=self.is_surface,
        )
        parent._master_component.part.bodies.append(tb)
        parent._clear_cached_bodies()
        body_id = f"{parent.id}/{tb.id}" if parent.parent_component else tb.id
        return Body(body_id, response.get("name"), parent, tb)

    @ensure_design_is_active
    def get_named_selections(self) -> list["NamedSelection"]:  # noqa: D102
        included_ns = []
        for ns in get_design_from_body(self).named_selections:
            if any(body.id == self.id for body in ns.bodies):
                included_ns.append(ns)

        return included_ns

    @ensure_design_is_active
    def get_raw_tessellation(  # noqa: D102
        self,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> dict:
        raw_tess = self._template.get_raw_tessellation(
            tess_options,
            reset_cache,
            include_faces,
            include_edges,
        )

        # Transform the raw tessellation points for both faces/edges
        from copy import deepcopy

        import numpy as np

        transform = self.parent_component.get_world_transform()
        transformed_map = deepcopy(raw_tess)
        for id, tess in raw_tess.items():
            vertices = np.reshape(np.array(tess.get("vertices")), (-1, 3))
            homogenous_points = np.hstack([vertices, np.ones((vertices.shape[0], 1))])
            transformed_points = (transform @ homogenous_points.T).T[:, :3]
            transformed_map[id]["vertices"] = transformed_points.flatten().tolist()

        return transformed_map

    @ensure_design_is_active
    def tessellate(  # noqa: D102
        self,
        merge: bool = False,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
    ) -> Union["PolyData", "MultiBlock"]:
        return self._template.tessellate(
            merge,
            self.parent_component.get_world_transform(),
            tess_options,
            reset_cache,
            include_faces,
            include_edges,
        )

    @graphics_required
    def get_vtk_tessellation(  # noqa: D102
        self,
        merge: bool = False,
        tess_options: TessellationOptions | None = None,
        reset_cache: bool = False,
        include_faces: bool = True,
        include_edges: bool = False,
        _raw_tessellation: dict | None = None,
    ) -> Union["vtkPolyData", "vtkMultiBlockDataSet"]:
        return self._template.get_vtk_tessellation(
            merge=merge,
            tess_options=tess_options,
            reset_cache=reset_cache,
            include_faces=include_faces,
            include_edges=include_edges,
            _raw_tessellation=_raw_tessellation
            or self.get_raw_tessellation(
                tess_options=tess_options,
                reset_cache=reset_cache,
                include_faces=include_faces,
                include_edges=include_edges,
            ),
        )

    @ensure_design_is_active
    def shell_body(self, offset: Distance | Quantity | Real) -> bool:  # noqa: D102
        return self._template.shell_body(offset)

    @ensure_design_is_active
    def remove_faces(  # noqa: D102
        self,
        selection: Face | Iterable[Face],
        offset: Distance | Quantity | Real,
    ) -> bool:
        return self._template.remove_faces(selection, offset)

    @graphics_required
    def plot(  # noqa: D102
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        # lazy import here to improve initial module load time
        from ansys.tools.visualization_interface.types.mesh_object_plot import (
            MeshObjectPlot,
        )

        import ansys.geometry.core as pyansys_geometry
        from ansys.geometry.core.plotting import GeometryPlotter

        use_service_colors = (
            use_service_colors
            if use_service_colors is not None
            else pyansys_geometry.USE_SERVICE_COLORS
        )

        # Add to plotting options as well... to be used by the plotter if necessary
        plotting_options["merge_bodies"] = merge

        # In case that service colors are requested, merge will be ignored.
        # An info message will be issued to the user.
        if use_service_colors and merge:
            self._template._grpc_client.log.info(
                "Ignoring 'merge' option since 'use_service_colors' is set to True."
            )
            plotting_options["merge_bodies"] = False

        mesh_object = (
            self if use_service_colors else MeshObjectPlot(self, self.tessellate(merge=merge))
        )
        pl = GeometryPlotter(use_trame=use_trame, use_service_colors=use_service_colors)
        pl.plot(mesh_object, **plotting_options)
        pl.show(screenshot=screenshot, **plotting_options)

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "intersect", "bodies do not intersect")
        else:
            self.__generic_boolean_command(
                other, keep_other, "intersect", "bodies do not intersect"
            )

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "subtract", "empty (complete) subtraction")
        else:
            self.__generic_boolean_command(
                other, keep_other, "subtract", "empty (complete) subtraction"
            )

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "unite", "union operation failed")
        else:
            self.__generic_boolean_command(other, False, "unite", "union operation failed")

    def combine_merge(self, other: Union["Body", list["Body"]]) -> None:  # noqa: D102
        self._template.combine_merge(other)

    @min_backend_version(27, 1, 0)
    def detach_faces(self) -> list["Body"]:
        """Detach all the faces from the body.

        This method will result in the original body
        becoming a surface body and a list of new surface bodies being created for every
        detached face.

        Returns
        -------
        list[Body]
            Bodies created by the detach if any.

        Warnings
        --------
        This method is only available starting on Ansys release 27R1.
        """
        response = self._grpc_client.services.model_tools.detach_faces(selections=[[self.id]])

        parent_design = get_design_from_body(self)

        if response.get("success"):
            if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
                parent_design._update_design_inplace()
            else:
                parent_design._update_from_tracker(response.get("tracked_response"))

            result_bodies = response.get("created_bodies")
            return get_bodies_from_ids(parent_design, result_bodies)
        else:
            self._grpc_client.log.info("Failed to detach faces.")
            return []

    @min_backend_version(26, 1, 0)
    def _combine_subtract(  # noqa: D102
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool = False,
        transfer_named_selections=True,
    ) -> None:
        parent_design = get_design_from_body(self)
        other = other if isinstance(other, Iterable) else [other]

        response = self._template._grpc_client.services.bodies.combine(
            target=self.id,
            other=[body.id for body in other],
            type_bool_op="subtract",
            keep_other=keep_other,
            transfer_named_selections=transfer_named_selections,
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response["tracker_response"])

    @reset_tessellation_cache
    @ensure_design_is_active
    @check_input_types
    def __generic_boolean_command(
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool,
        method: str,
        err_msg: str,
    ) -> None:
        parent_design = get_design_from_body(self)
        other = other if isinstance(other, Iterable) else [other]

        response = self._template._grpc_client.services.bodies.combine(
            target=self.id,
            other=[body.id for body in other],
            type_bool_op=method,
            err_msg=err_msg,
            keep_other=keep_other,
            transfer_named_selections=False,
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response["tracker_response"])

    @reset_tessellation_cache
    @ensure_design_is_active
    @check_input_types
    def __generic_boolean_op(
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool,
        method: str,
        err_msg: str,
    ) -> None:
        grpc_other = other if isinstance(other, Iterable) else [other]
        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN and keep_other:
            # Make a copy of the other body to keep it...
            # stored temporarily in the parent component - since it will be deleted
            grpc_other = [b.copy(self.parent_component, f"BoolOpCopy_{b.name}") for b in grpc_other]

        response = self._template._grpc_client.services.bodies.boolean(
            target=self.id,
            other=[other.id for other in grpc_other],
            method=method,
            err_msg=err_msg,
            keep_other=keep_other,
        )

        parent_design = get_design_from_body(self)
        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            for b in grpc_other:
                b.parent_component.delete_body(b)
            parent_design._update_design_inplace()
        else:
            # If USE_TRACKER_TO_UPDATE_DESIGN is True, we serialize the response
            # and update the parent design with the serialized response.
            parent_design._update_from_tracker(response["tracker_response"])

    def __repr__(self) -> str:
        """Represent the ``Body`` as a string."""
        lines = [f"ansys.geometry.core.designer.Body {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Parent component     : {self._parent_component.name}")
        lines.append(f"  MasterBody           : {self._template.id}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")
        lines.append(f"  Color                : {self.color}")

        nl = "\n"
        return f"{nl}{nl.join(lines)}{nl}"
